

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <map>
#include <iterator>
#include "plink.h"
#include "options.h"

namespace std
{
  template<>
  class less<Segment*> {
  public:
  bool operator()(Segment const* s1, Segment const* s2)
  {
    if      ( s1->start  > s2->start ) return true;
    else if ( s1->start  < s2->start ) return false;
    else if ( s1->finish > s2->finish ) return true;
    else if ( s1->p1 < s2->p1 ) return true;
    else if ( s1->p1 > s2->p1 ) return false;
    else if ( s1->p2 < s2->p2 ) return true;
    else if ( s1->p2 > s2->p2 ) return false;
    else return false;
  }
  };
};

class Pool
{
public:    
  set<Segment*,less<Segment*> > segs;
  vector<vector<int> > match;
  vector<int> matchcount;
  vector<int> group;
  vector<bool> index;
  int ng;
  int min;
  int max;
  int union_min;
  int union_max;
};

namespace std
{
  template<>
  class less<Pool*> {
  public:
  bool operator()(Pool const* p1, Pool const* p2)
  {

    if      ( p1->segs.size() > p2->segs.size() ) return true; 
    else if ( p1->segs.size() < p2->segs.size() ) return false; 
    else 
      {
	set<Segment*>::iterator s1 = p1->segs.begin();
	set<Segment*>::iterator s2 = p2->segs.begin();
	
	while ( s1 != p1->segs.end() )
	  {
	    if      ( (*s1)->start > (*s2)->start ) return true; 
	    else if ( (*s1)->start < (*s2)->start ) return false; 
	    else if ( (*s1)->finish > (*s2)->finish ) return true; 
	    else if ( (*s1)->finish < (*s2)->finish ) return false; 
	    s1++;
	    s2++;
	  }
      }
    return false;
  }
  };
};



// Homozygosity match, whole segments 
bool segsMatch(Segment * s1, Segment * s2)
{

  int mismatch = 0;
  
  int start = s1->start > s2->start ? s1->start : s2->start;
  int finish = s1->finish < s2->finish ? s1->finish : s2->finish;
  
  // Assume individual major mode
  
  for (int l=start; l<= finish; l++)
    {
      // Only test homozygotes
      if ( s1->p1->one[l] == s1->p1->two[l] && 
	   s2->p1->one[l] == s2->p1->two[l] )
	if ( s1->p1->one[l] != s2->p1->one[l] )
	  mismatch++;
    }

  if ((double)mismatch/(double)(finish-start+1) > 1-par::fuzzy_homo ) return false;
  else return true;
}


// Homozygosity Match, supplying pool consensus start/finish
bool segsIBDMatchCON(Segment * s1, Segment * s2, int start, int finish)
{
  int match = 0;
  int valid = 0;

  // 1. Determine which allele is shared for each pair
  // 2. See whether this is the same allele between pairs 
 
  // Assume individual major mode
    for (int l=start; l<= finish; l++)
    {

      // Ignore is 1+ missing or 2 hets
      
      bool a1 = s1->p1->one[l];
      bool a2 = s1->p1->two[l];

      bool b1 = s1->p2->one[l];
      bool b2 = s1->p2->two[l];
      
      bool c1 = s2->p1->one[l];
      bool c2 = s2->p1->two[l];

      bool d1 = s2->p2->one[l];
      bool d2 = s2->p2->two[l];

      bool allele1 = false;
      bool allele2 = false;

      // Any missing alleles?
      if ( (a1 && (!a2)) || 
	   (b1 && (!b2)) || 
	   (c1 && (!c2)) ||
	   (d1 && (!d2)) )
	continue;

      // Any double hets within each pair?
      if ( ((!a1) && a2) &&
	   ((!b1) && b2) )
	continue;
	 
      if ( (!c1) && c2 &&
	   (!d1) && d2 )
	continue;

      // Any opposing homozygotes within pair?
      if ( (a1 == a2) &&
	   (b1 == b2) && 
	   (a1 != b1) )
	continue;

      if ( (c1 == c2) &&
	   (d1 == d2) && 
	   (c1 != d1) )
	continue;
      
      // Get the alleles for the pairs
      // (i.e. from the homozygote)

      if ( a1 == a2 ) allele1 = a1;
      else allele1 = b1;
      
      if ( c1 == c2 ) allele2 = c1;
      else allele2 = d1;

      // Do these match?
      if ( allele1 == allele2 ) 
	match++;
      
      valid++;
    }


  if ( valid<par::genome_test_min_snp ) return false;
  if ((double)match/(double)valid < par::fuzzy_homo ) return false;
  else return true;


}


 // Pairwise segmental match, whole segments 
 bool segsIBDMatch(Segment * s1, Segment * s2)
{

  int match = 0;
  int valid = 0;

  int start = s1->start > s2->start ? s1->start : s2->start;
  int finish = s1->finish < s2->finish ? s1->finish : s2->finish;

   // 1. Determine which allele is shared for each pair
  // 2. See whether this is the same allele between pairs 
 
  // Assume individual major mode
  
  for (int l=start; l<= finish; l++)
    {

      // Ignore is 1+ missing or 2 hets
      
      bool a1 = s1->p1->one[l];
      bool a2 = s1->p1->two[l];

      bool b1 = s1->p2->one[l];
      bool b2 = s1->p2->two[l];
      
      bool c1 = s2->p1->one[l];
      bool c2 = s2->p1->two[l];

      bool d1 = s2->p2->one[l];
      bool d2 = s2->p2->two[l];

      bool allele1 = false;
      bool allele2 = false;

      // Any missing alleles?
      if ( (a1 && (!a2)) || 
	   (b1 && (!b2)) || 
	   (c1 && (!c2)) ||
	   (d1 && (!d2)) )
	continue;

      // Any double hets within each pair?
      if ( ((!a1) && a2) &&
	   ((!b1) && b2) )
	continue;
	 
      if ( (!c1) && c2 &&
	   (!d1) && d2 )
	continue;
   
      // Any opposing homozygotes within pair?
      if ( (a1 == a2) &&
	   (b1 == b2) && 
	   (a1 != b1) )
	continue;

      if ( (c1 == c2) &&
	   (d1 == d2) && 
	   (c1 != d1) )
	continue;

      // Get the alleles for the pairs
      // (i.e. from the homozygote)

      if ( a1 == a2 ) allele1 = a1;
      else allele1 = b1;
      
      if ( c1 == c2 ) allele2 = c1;
      else allele2 = d1;

      // Do these match?
      if ( allele1 == allele2 ) 
	match++;
      
      valid++;
    }


  if ( valid<par::genome_test_min_snp ) return false;
  if ((double)match/(double)valid < par::fuzzy_homo ) return false;
  else return true;
}


// Pairwise segmental match, supplying pool consensus start/finish
bool segsMatchCON(Segment * s1, Segment * s2, int start, int finish)
{

  int mismatch = 0;
  
  // Assume individual major mode
  
  for (int l=start; l<= finish; l++)
    {
      // Only test homozygotes
      if ( s1->p1->one[l] == s1->p1->two[l] && 
	   s2->p1->one[l] == s2->p1->two[l] )
	if ( s1->p1->one[l] != s2->p1->one[l] )
	  mismatch++;
    }

  if ((double)mismatch/(double)(finish-start+1) > 1-par::fuzzy_homo ) return false;
  else return true;
}


void displayPoolVerbose( Plink & P, Pool * pool , ofstream & OUT )
{
  
  // Figure out list of individuals, for the maximal region 
  // i.e. union rather than intersection of the pool of segs

  set<Individual*> pset;
  vector<Individual*> plist;
  vector<int> pstart;
  vector<int> pend;
  vector<int> pgrp;

  // Loop over each group in the pool
  for ( int g = 0; g < pool->ng; g++)
    {
      
      // Consider all segments in this pool
      set<Segment*>::iterator s = pool->segs.begin();
      int c2=0;
      while ( s != pool->segs.end() )
	{

	  // Not in group 'g' ?
	  if ( pool->group[c2] == g ) 
	    {
	      
	      if ( pset.find( (*s)->p1 ) == pset.end() )
		{
		  pset.insert( (*s)->p1 );
		  plist.push_back( (*s)->p1 );
		  pstart.push_back( (*s)->start );
		  pend.push_back( (*s)->finish );
		  pgrp.push_back( g );
		}
	      
	      if ( pset.find( (*s)->p2 ) == pset.end() )
		{
		  pset.insert( (*s)->p2 );
		  plist.push_back( (*s)->p2 );
		  pstart.push_back( (*s)->start );
		  pend.push_back( (*s)->finish );
		  pgrp.push_back( g );
		}
	      
	    }
	  
	  // Next segment
	  s++;	  
	  c2++;
	  continue;

	}
    }

  // We now a have a unique list of individuals, in the same order in plist as
  // the grouping variable; now display, each row is a SNP

  // Header row (including IDs)

  OUT << setw(par::pp_maxsnp) << "SNP" << " ";
  
  int fsize = par::pp_maxfid > par::pp_maxiid ? par::pp_maxfid : par::pp_maxiid;
  for (int i=0; i < plist.size(); i++)
    OUT << setw(fsize) << plist[i]->fid << " ";
  OUT << "\n" << setw(par::pp_maxsnp) << " " << " ";
 
  for (int i=0; i < plist.size(); i++)
    OUT << setw(fsize) << plist[i]->iid << " ";
  OUT << "\n" << setw(par::pp_maxsnp) << " " << " ";
  for (int i=0; i < plist.size(); i++)
    OUT << setw(fsize) << plist[i]->iid << " ";
  OUT << "\n" << setw(par::pp_maxsnp) << " " << " ";
  for (int i=0; i < plist.size(); i++)
    OUT << setw(fsize) << pgrp[i] << " ";
  OUT << "\n\n";

  for (int l = pool->union_min ; l <= pool->union_max ; l++)
    {
      if ( l == pool->min ) OUT << "\n";

      OUT << setw(par::pp_maxsnp) << P.locus[l]->name << " ";
      
      for (int i=0; i < plist.size(); i++)
	{
	  if ( l >= pstart[i] && l <= pend[i] )
	    OUT << setw(fsize) << "["+genotype(P,plist[i],l)+"]" << " ";
	  else
	    OUT << setw(fsize) << " "+genotype(P,plist[i],l)+" " << " ";
	}
      OUT << "\n";

      if ( l == pool->max ) OUT << "\n";

    }

  OUT << "\n\n";
}


void Plink::summariseHomoRuns()
{

  string f;
  if (par::segment_overlap)
    f = par::output_file_name + ".segment.overlap";
  else
    f = par::output_file_name + ".hom.overlap";

  ofstream HOM(f.c_str(),ios::out);
  HOM.precision(6);

  if (par::segment_overlap)
    printLOG("Writing IBD-run summary to [ "+f+" ] \n");
  else
    printLOG("Writing homozygosity-run summary to [ "+f+" ] \n");
  
  
  if (par::fuzzy_homo > 1e-6) 
    printLOG("Requiring allelic match, threshold "+dbl2str(par::fuzzy_homo)+" identity\n");
  else
    {
      printLOG("Not requiring allelic match (overlapping only)\n");
      par::fuzzy_homo = 0;
    }

  if (par::segment_overlap)
    HOM << setw(par::pp_maxfid) << "FID1" << " "
	<< setw(par::pp_maxiid) << "IID1" << " "
	<< setw(par::pp_maxfid) << "FID2" << " "
	<< setw(par::pp_maxiid) << "IID2" << " "
	<< setw(8) << "PHE" << " "
	<< setw(4) << "CHR" << " " 
	<< setw(par::pp_maxsnp) << "SNP1" << " " 
	<< setw(par::pp_maxsnp) << "SNP2" << " "
	<< setw(14) << "BP1" << " " 
	<< setw(14) << "BP2" << " "
	<< setw(8) << "KB" << " "
	<< setw(4) << "NS" << " " 
	<< setw(6) << "GRP" << "\n";
  else
    HOM << setw(par::pp_maxfid) << "FID" << " "
	<< setw(par::pp_maxiid) << "IID" << " "
	<< setw(8) << "PHE" << " "
	<< setw(4) << "CHR" << " " 
	<< setw(par::pp_maxsnp) << "SNP1" << " " 
	<< setw(par::pp_maxsnp) << "SNP2" << " "
	<< setw(14) << "BP1" << " " 
	<< setw(14) << "BP2" << " "
	<< setw(8) << "KB" << " "
  	<< setw(4) << "NS" << " " 
	<< setw(6) << "GRP" << "\n";

  // Make groups of segments
  // Group by maximally over-lapping region
  
  // Both irrespective of alleles
  // And considering alleles
  
  // i.e. all segments in a group must have at least one position where they overlap
  // Then group alleles within pool based on allelic identity
  
  //   ----                  -----
  // ----                 ----------
  //   -----------------------
  //   |                     |




  // A unique list of overlapping segments
  set<Pool*,less<Pool*> > pools;

  ///////////////
  // 1. Make sets
  
  // Consider each position
  int cons=0;
  for (int l=0; l<nl_all; l++)
    {
      
      if (!par::silent)
	cout << "Considering position " << l+1 << " of " << nl_all << "        \r";

      Pool * thispool = new Pool;
      
      // Which segments contain this SNP?
      vector<Segment>::iterator s = segment.begin();      
      while ( s != segment.end() )
	{
	  if ( s->start <= l && 
	       s->finish >= l )
	    {
	      thispool->segs.insert( &(*s) );
	    }
	  s++;
	}


      // Add this pool to the overall list, if it is unique
      if ( thispool->segs.size()>1 )
	{
	  pools.insert( thispool );
	}
      else delete thispool;
    }

  if (!par::silent)
    cout << "\n";

  printLOG("Found "+int2str(pools.size())+" unique overlapping sets of segments\n");
  

  //////////////////////////////////////////////////////
  // 2. Prune,  i.e.  drop AB if ABC and/or ABD exists?
  // Start at end (smallest set) and go upwards
  
  vector<bool> redundant(pools.size(),false);
  
  set<Pool*>::reverse_iterator p = pools.rbegin();
  int c=pools.size()-1;
  while ( p != pools.rend()) 
    {

      if (!par::silent)
	cout << c << " pools left to prune    \r";

      // Do we find all segments in this pool in a larger pool?
      
      set<Pool*>::iterator p2 = pools.begin();
      int c2=0;
      while ( p2 != pools.end() )
 	{
	  
 	  // Already redundant
 	  if ( c2 >= c || redundant[c2] )
 	    {
 	      c2++;
 	      p2++;
 	      continue;
 	    }
	  
 	  // Same pool
 	  if ( *p == *p2 )
 	    {
 	      c2++;
 	      p2++;
 	      continue;
 	    }
	  
 	  set<Segment*>::iterator s = (*p)->segs.begin();
 	  bool embedded = true;
	  
 	  while ( s != (*p)->segs.end() )
 	    {
 	      // Do we find this segment?
	      
 	      set<Segment*>::iterator s2 = (*p2)->segs.begin();
 	      bool found = false;
	      
 	      while ( s2 != (*p2)->segs.end() )
 		{
 		  if ( *s == *s2 ) 
 		    {
 		      found = true;
 		      break;
 		    }
 		  s2++;
 		}
	      
 	      if ( ! found ) embedded = false;
	      
 	      s++;
	    }
	  
 	  if (embedded) 
 	    {
 	      redundant[c] = true;
 	      break;
	    }

	  c2++;
 	  p2++;
 	}      
      c--;
      p++;
    }
  

  if (!par::silent)
    cout << "\n";


  //////////////////////////////////////////////
  // 3. Determine allelic matching

 
  // Determine consensus region for each pool
  
  set<Pool*>::iterator pC = pools.begin();
  while ( pC != pools.end()) 
    {
      
      (*pC)->min=0;
      (*pC)->max=nl_all;

      (*pC)->union_min=nl_all;
      (*pC)->union_max=0;

      set<Segment*>::iterator s = (*pC)->segs.begin();
      while ( s != (*pC)->segs.end() )
	{
	  (*pC)->min = (*s)->start > (*pC)->min ? (*s)->start : (*pC)->min;
 	  (*pC)->max = (*s)->finish < (*pC)->max ? (*s)->finish : (*pC)->max;

	  (*pC)->union_min = (*s)->start < (*pC)->union_min ? 
	    (*s)->start : (*pC)->union_min;
 	  (*pC)->union_max = (*s)->finish > (*pC)->union_max ? 
	    (*s)->finish : (*pC)->union_max;

	  s++;

	}  

      pC++;
    }

  // Find matches
 
  set<Pool*>::iterator pA = pools.begin();
  int cA=0;
  while ( pA != pools.end()) 
    {

      // Skip this pool?
      if ( redundant[cA] )
	{
 	  cA++;
 	  pA++;
 	  continue;
	}
      
      // For this pool, consider all segments in the pool in a pairwise
      // manner, and populate the match matrix

      set<Segment*>::iterator s1 = (*pA)->segs.begin();
      int c1=0;
      
      // Resize matching variables
      (*pA)->match.resize( (*pA)->segs.size() );
      for (int i=0; i< (*pA)->segs.size(); i++) 
	(*pA)->match[i].clear();
      (*pA)->group.resize( (*pA)->segs.size() , 0 );

      (*pA)->matchcount.resize((*pA)->segs.size() , 0 );
      (*pA)->index.resize((*pA)->segs.size() , false );
	    
      // Conisder each pair of segments
      while ( s1 != (*pA)->segs.end() )
	{

	  set<Segment*>::iterator s2 = (*pA)->segs.begin();
	  int c2=0;
	  
	  while ( s2 != (*pA)->segs.end() )
	    {
	      
	      if ( c2 >= c1 )
		{
		  s2++;
		  c2++;
		  continue;
		}
	      
	      // Determine match function:
	      // Based on homozygosity or pairwise sharing
	      // Based on whole segments, or just pool consensus region
	      
	      if (par::segment_overlap)
		{
		  // PAIRWISE SEGMENTAL MATCH

 		  if ( par::homo_run_consensus_match)
 		    { 
 		      // Consensus match
 		      if ( segsIBDMatchCON( *s1, *s2, (*pA)->min, (*pA)->max ) )
 			{
 			  (*pA)->match[c1].push_back( c2 );
 			  (*pA)->match[c2].push_back( c1 );		  
 			  (*pA)->matchcount[c1]++;
 			  (*pA)->matchcount[c2]++;
 			}
 		    }
 		  else // else match whole segments (default) 
 		    {		      
 		      if ( segsIBDMatch( *s1, *s2 ) )
 			{
 			  (*pA)->match[c1].push_back( c2 );
 			  (*pA)->match[c2].push_back( c1 );		  
 			  (*pA)->matchcount[c1]++;
 			  (*pA)->matchcount[c2]++;
 			}
		      
 		    }
		}
	      else
		{
		  // HOMOZYGOSITY MATCH
		  // If a match, add pairs to lists
		  if ( par::homo_run_consensus_match)
		    { 
		      // Consensus match
		      if ( segsMatchCON( *s1, *s2, (*pA)->min, (*pA)->max ) )
			{
			  (*pA)->match[c1].push_back( c2 );
			  (*pA)->match[c2].push_back( c1 );		  
			  (*pA)->matchcount[c1]++;
			  (*pA)->matchcount[c2]++;
			}
		    }
		  else // else match whole segments (default) 
		    {
		      if ( segsMatch( *s1, *s2 ) )
			{
			  (*pA)->match[c1].push_back( c2 );
			  (*pA)->match[c2].push_back( c1 );		  
			  (*pA)->matchcount[c1]++;
			  (*pA)->matchcount[c2]++;
			}
		      
		    }
		}
	      
	      // Next segment (B)
	      s2++;
	      c2++;
	    }
	  
	  // Next segment (A)
	  s1++;
	  c1++;
	}


      // Parse the list
      bool done = false;
      (*pA)->ng = 1;

      while ( ! done )
	{

	  // Find largest, ungrouped list
	  int maxlist = 0;
	  int maxlisti = -1;
	  
	  for (int i=0; i < (*pA)->group.size() ; i++)
	    {
	      if ( (*pA)->group[i] == 0 ) 
		{ 
		  if ( (*pA)->match[i].size() >= maxlist ) 
		    {
		      maxlist = (*pA)->match[i].size();
		      maxlisti = i;
		    } 
		}
	    }

	  // Set group for this, and matches
	  (*pA)->index[maxlisti] = true;
	  (*pA)->group[maxlisti] = (*pA)->ng;
	  for ( int j=0; j < maxlist; j++)
	    (*pA)->group[ (*pA)->match[maxlisti][j] ] = (*pA)->ng;

	  // Advance to next group
	  (*pA)->ng++;

	  // Are all segments grouped?
	  bool ungroup = false;
	  
	  for (int i=0; i < (*pA)->group.size() ; i++)
	    {
	      if ( (*pA)->group[i] == 0 ) 
		{
		  ungroup = true;
		  //break;
		}
	    }
	  
	  if ( ! ungroup ) done = true;
	  
	} 


      // Remove temporary storage
      (*pA)->match.clear();

      // Next pool
      pA++;
      cA++;
    }
  
  
  //////////////
  // 4. Display

  set<Pool*>::iterator p2 = pools.begin();
  c=0;
  while ( p2 != pools.end()) 
    {
      
      if (redundant[c])
	{
 	  c++;
  	  p2++;
  	  continue;
	}

      int ncase=0;
      int ncontrol=0;
      
      // Loop over each group in the pool
      for ( int g = 0; g < (*p2)->ng; g++)
	{

	  // Consider all segments in this pool
	  set<Segment*>::iterator s = (*p2)->segs.begin();
	  int c2=0;
	  while ( s != (*p2)->segs.end() )
	    {

	      // Not in group 'g' ?
	      if ( (*p2)->group[c2] != g ) 
		{
		  s++;
		  c2++;
		  continue;
		} 
	      
	      if (par::segment_overlap)
		{
		  HOM << setw(par::pp_maxfid) << (*s)->p1->fid << " "
		      << setw(par::pp_maxiid) << (*s)->p1->iid << " "
		      << setw(par::pp_maxfid) << (*s)->p2->fid << " "
		      << setw(par::pp_maxiid) << (*s)->p2->iid << " ";
		  if (par::bt)
		    {
		      if ( (*s)->p1->missing || (*s)->p2->missing )
			HOM << setw(8) << "NA" << " ";
		      else if ( (!(*s)->p1->aff) && (!(*s)->p2->aff) ) 
			{ 
			  HOM << setw(8) << "-1" << " ";
			  ncontrol++;
			}
		      else if ( (*s)->p1->aff && (*s)->p2->aff )
			{
			  HOM << setw(8) << "1" << " ";
			  ncase++;
			}
		      else 
			{
			  HOM << setw(8) << "0" << " ";
			  ncontrol++;
			}
		    }
		  else
		    HOM << setw(8) << "NA" << " ";
		}
	      else
		{
		  HOM << setw(par::pp_maxfid) << (*s)->p1->fid << " "
		      << setw(par::pp_maxiid) << (*s)->p1->iid << " ";
		  
		  if (par::bt)
		    HOM << setw(8) << (*s)->p1->phenotype << " ";
		  else
		    {
		      HOM.precision(4);
		      HOM << setw(8) << (*s)->p1->phenotype << " ";		      
		      HOM.precision(8);
		    }

		  if ( (*s)->p1->aff ) 
		    ncase++;
		  else
		    ncontrol++;
		}
	      
	      HOM << setw(4) << locus[(*s)->start]->chr << " "
		  << setw(par::pp_maxsnp) << locus[(*s)->start]->name << " "
		  << setw(par::pp_maxsnp) << locus[(*s)->finish]->name << " "
		  << setw(14) << locus[(*s)->start]->bp << " "
		  << setw(14) << locus[(*s)->finish]->bp << " "
		  << setw(8) << (double)(locus[(*s)->finish]->bp - locus[(*s)->start]->bp ) / 1000.0 
		  << setw(4) << (*p2)->matchcount[c2] << " ";
	      if ( (*p2)->index[c2] )
		HOM << setw(6) << int2str((*p2)->group[c2])+"*" << " ";
	      else
		HOM << setw(6) << int2str((*p2)->group[c2])+" " << " ";
	      
	      HOM << "\n";
	      	      
	      s++;
	      c2++;

	    } // Next segment 
	  
	} // Next group
      
      // Consensus region      
      
      HOM << setw(par::pp_maxfid) << "CON" << " "
	  << setw(par::pp_maxiid) <<  (*p2)->segs.size() << " ";
      if (par::segment_overlap)
	HOM << setw(par::pp_maxfid) << "NA" << " "
	    << setw(par::pp_maxiid) << "NA" << " ";
      HOM << setw(8) << int2str(ncase)+":"+int2str(ncontrol) << " "
	  << setw(4) << locus[(*(*p2)->segs.begin())->start]->chr << " "
	  << setw(par::pp_maxsnp) << locus[(*p2)->min]->name << " "
	  << setw(par::pp_maxsnp) << locus[(*p2)->max]->name << " "
	  << setw(14) << locus[(*p2)->min]->bp << " "
	  << setw(14) << locus[(*p2)->max]->bp << " "
	  << setw(8) << (double)(locus[(*p2)->max]->bp - locus[(*p2)->min]->bp ) / 1000.0 ;
      HOM << "\n\n";

      
      // Verbose mode?
      
      if (par::homozyg_verbose || par::segment_verbose)
	displayPoolVerbose( *this , *p2 , HOM );

      c++;
      p2++;
      
    } // Next pool
  
  
  
  HOM.close();  
}




void Plink::findHomoRuns(Individual * person, ofstream & HOM)
{

  int l=0; 
  int lasthom=0;
  int nmiss = 0;
  int nhet = 0;
  bool run = false;
  int start = 0;
  int end = 0;

  while ( l < nl_all )
    {
      
      // Skip haploid chromosomes / end any existing run
      if ( ( par::chr_sex[locus[l]->chr] && person->sex ) || 
	   par::chr_haploid[locus[l]->chr] ) 
	{
	  if (run) 
	    {
	      end = l-1;
	      run = false;
	    }
	  else
	    {
	      l++;
	      continue; 
	    }
	}
      
      // Outside of a run?
      if (!run)
	{
	  // A new run?
	  if (person->one[l] == person->two[l])
	    {
	      start = lasthom = l;
	      nmiss=0;
	      nhet=0;
	      run=true;
	    }
	}
      else // if already in a run, either end or increase length?
	{

	  if ( locus[l]->chr != locus[start]->chr ) // different chromosome?
	    {
	      end = l-1;
	      run = false;
	    }
	  else if ( l == (nl_all -1) ) // or end of all SNPs?
	    {
	      if ( person->one[l] == person->two[l] )
		lasthom=l;
	      end = lasthom;
	      run = false;
	    }
	  // found a het?
	  else if ( (!person->one[l]) && person->two[l])
	    {
	      if (nhet==par::homo_run_het)
		{
		  end = lasthom;
		  run = false;
		}
	      else
		nhet++;
	    }
	  else // ...continue run
	    {
	      lasthom=l;
	      if ( person->one[l] && (!person->two[l]) ) 
		nmiss++;
	    }
	  
	}


      // Check run length?
      if (!run)
	{

	  bool accept = true;
	  
	  if (par::homo_run_kb) 
	    if ( locus[end]->bp - locus[start]->bp < 
		 par::homo_run_length_kb * 1000 )
	      accept = false;
	  
	  if (par::homo_run_snps)
	    if ( end - start +1 < par::homo_run_length_snps  )
	      accept = false;
	  
	  if (accept)
	    {
	      HOM << setw(par::pp_maxfid) << person->fid << " "
		  << setw(par::pp_maxiid) << person->iid << " "
		  << setw(4) << locus[start]->chr << " "
		  << setw(par::pp_maxsnp) << locus[start]->name << " "
		  << setw(par::pp_maxsnp) << locus[end]->name << " "
		  << setw(12) <<locus[start]->bp << " "
		  << setw(12) <<locus[end]->bp << " "
		  << setw(10) << (double)(locus[end]->bp - locus[start]->bp)/(double)1000 << " "
		  << setw(10) <<end - start + 1 << " "
		  << setw(4) << nhet << " "
		  << setw(4) << nmiss << "\n";
	      
	      Segment s;
	      s.p1 = s.p2 = person;
	      s.start = start;
	      s.finish = end;
	      segment.push_back(s);
	    }
	  
	  
	  //////////////////
	  // Clear counters

	  start = end = nmiss = 0;
	  
	}
      
      ///////////////
      // Next locus

      l++;
    }

}



bool segsOverlap(Segment * s1, Segment * s2)
{
  if ( s1->finish < s2->start ) return false;
  else if ( s2->finish < s1->start ) return false;
  else return true;  
}




