

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include "model.h"
#include "options.h"
#include <cmath>

Model::Model()
{
  np = nind = 0;
  
  testParameter = 1; // Permutation test parameter
}

void Model::setMissing()
{
  
  // Fill in missing data with existing pattern 
  
  miss.resize(P->n,false);
  for (int i=0; i<P->n; i++)
    if ( P->sample[i]->missing ) miss[i] = true;
  
}



void Model::addAdditiveSNP(CSNP * a)
{
  additive.push_back(a);
}

void Model::addDominanceSNP(CSNP * d)
{
  dominance.push_back(d);
}

void Model::addCovariate(int c)
{
  covariate.push_back(c);      
}

void Model::addInteraction(int a, int b)
{
  int2 i;
  i.p1 = a;
  i.p2 = b;
  interaction.push_back(i);
}

void Model::buildDesignMatrix()
{
  

  // Assumptions: 
  // that missing and Y have already been set, with nind
  // np = 1 currently (intercept)

  

  /////////////////////////////
  // 1) Main effects of SNPs 
  // 2) Dominance effects of SNPs 
  // 3) Covariates
  // 4) Interactions of the above

  
  ///////////////////
  // Additive effects

  vector<CSNP*>::iterator s = additive.begin();
  
  while ( s != additive.end() )
    {
  
      vector<bool>::iterator i1 = (*s)->one.begin();
      vector<bool>::iterator i2 = (*s)->two.begin();
      int c=0;
      
      while ( i1 != (*s)->one.end() )
	{
	  
	  // SNP alleles	  
	  	  
	  if ( *i1 ) 
	    {
	      if ( ! (*i2) ) 
		{
		  miss[c] = true;
		  X[c].push_back(0);
		}
	      else
		{
		  X[c].push_back(0);
		}
	    }
	  else
	    {
	      if ( * i2 )
		X[c].push_back(1); // het
	      else
		X[c].push_back(2); // hom
	    }
	  
	  // Advance to the next person	  
	  i1++;
	  i2++;
	  c++;	  
	}
      
      // Next SNP / indepedent variable
      np++;
      s++;
    }
  
  
  ////////////////////
  // Dominance effects
  
  s = dominance.begin();
  
  while ( s != dominance.end() )
    {
      
      vector<bool>::iterator i1 = (*s)->one.begin();
      vector<bool>::iterator i2 = (*s)->two.begin();
      int c=0;
      
      while ( i1 != (*s)->one.end() )
	{
	  
	  // SNP alleles	  
	  
	  if ( *i1 ) 
	    {
	      if ( ! (*i2) ) 
		{
		  miss[c] = true;
		  X[c].push_back(0);
		}
	      else
		{
		  X[c].push_back(0);
		}
	    }
	  else
	    {
	      if ( * i2 )
		X[c].push_back(1); // het
	      else
		X[c].push_back(0); // hom
	    }

	  // Advance to the next person	  
	  i1++;
	  i2++;
	  c++;	  
	}
      
      // Next SNP / indepedent variable
      np++;
      s++;
    }


  /////////////
  // Covariates
  
  for (int c=0; c<covariate.size(); c++)
    {
      // Fill in information for all individuals
      // Missing values will already have been taken
      // Care of

      for (int i=0; i<P->n; i++)
	{
	  if ( ! miss[i] )
	    X[i].push_back( P->sample[i]->clist[covariate[c]] );
	  else
	    X[i].push_back( 0 );
	}

      // Add a new covariate
      np++;
    }


  ///////////////
  // Interactions
  
  for (int c=0; c<interaction.size(); c++)
    {
      
      // Intercept if X column 0, then subsequently added 
      // effects are in this order
      
      // Fill in information for all individuals
      for (int i=0; i<P->n; i++)
	X[i].push_back( X[i][interaction[c].p1] * X[i][interaction[c].p2] );
      
      // Add a new covariate
      np++;
    }
  
 
  /////////////////////////////////////////////////
  // Prune out rows that are missing, then set nind

  nind = 0;
  matrix_t X2;
  for (int i=0; i<X.size(); i++)
    if ( ! miss[i] ) 
      {
	X2.push_back(X[i]);
	nind++;
      }
  X = X2;

  // Now we are ready to perform the analysis
  
}


vector<bool> Model::validParameters()
{  

  // Check for multicollinearity
  
  // For each term, see that estimate is not too strongly (r>0.99)
  // correlated with another, starting at last
  
  valid.resize(np,true);
  all_valid = true;

  for (int i = np-1; i>0; i--)
    {
      for (int j = i-1; j>=0; j--)
	{
	  if ( S[i][j] / sqrt( S[i][i] * S[j][j] ) > 0.999 )
	    {
	      valid[i] = false;
	      all_valid = false;
	      break;
	    }
	}
    }
  return valid;
}


double Model::getStatistic()
{

  bool okay = S[testParameter][testParameter] < 1e-6 ? false : true;
  
  double se = 0; 
  double Z = 0;
  
  if (okay)
    {
      se = sqrt(S[testParameter][testParameter]);
      Z = coef[testParameter] / se;
    }
  
  return Z*Z;

}


// **********************************************
// *** Function to aid testing linear 
// *** hypotheses after estimating of a
// *** regression
// **********************************************

double Model::linearHypothesis(matrix_t & H, vector_t & h)
{
  
  //   return ( H*b - h ).transpose() 
  //     * ( H*v*H.transpose() ).inverse()
  //     * ( H*b - h ) ;
  
  int nc = h.size(); // # of constraints
  
  // 1. Calculate Hb-h  

  vector_t outer;
  outer.resize(nc,0);
  
  for (int r = 0; r < nc; r++)
    for (int c = 0; c < np; c++)
      outer[r] += H[r][c] * coef[c];
  
  for (int r = 0; r < nc; r++)
    outer[r] -= h[r];
  
  // 2. Calculate HVH'
  
  matrix_t tmp;
  sizeMatrix(tmp,nc,np);
  
  for (int r = 0; r < nc; r++)
    for (int c = 0; c < np; c++)
      for (int k = 0; k < np; k++)
	tmp[r][c] += H[r][k] * S[k][c];
  
  matrix_t inner;
  sizeMatrix(inner,nc,nc);
  
  for (int r = 0; r < nc; r++)
    for (int c = 0; c < nc; c++)
      for (int k = 0; k < np; k++)
	inner[r][c] += tmp[r][k] * H[c][k];

  inner = svd_inverse(inner);

  vector_t tmp2;
  tmp2.resize(nc,0);
  
  for (int c = 0; c < nc; c++)
    for (int k = 0; k < nc; k++)
      tmp2[c] += outer[k] * inner[k][c];

  double result = 0;
  
  for (int r = 0; r < nc; r++)
    result += tmp2[r] * outer[r];

  return result;
    
}
